/**
  Lists the dpes required for building the state of the device
  
  @param string sDeviceName The name of the device which is in UNICOS the same as the dpname
  @param dyn_string dsDpes The list of datapoints elements required for building the state
  @param bool bForHistory The DPEs will be used for state history, i.e. only DPEs with _archive
                            config are of interest
  
*/
vacuumExtension_StateDPEs(string sDeviceName, dyn_string &dsDpes, bool bForHistory = true) {
  // Load the front end name from the device name
  string frontEnd = unGenericObject_GetPLCNameFromDpName(sDeviceName);
  if(frontEnd != "") {
    string  systemName = unGenericDpFunctions_getSystemName(sDeviceName);
    frontEnd = systemName + c_unSystemAlarm_dpPattern + PLC_DS_pattern + frontEnd;
  }
  dynAppend(dsDpes, frontEnd+".alarm");
  dynAppend(dsDpes, frontEnd+".enabled");
  
  // Get the device name from the 
  string sDeviceType = dpTypeName(sDeviceName), sFunction = "";
  if(sDeviceType != ""){
    sFunction = sDeviceType +"_StateDPEs";
  } else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateDPEs: Device with name %s does not exists", sDeviceName);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 104, sErrMsg);
    throw( error );	
  }
  
  if (isFunctionDefined(sFunction) && sFunction != "") {
    // Execute callback function for the specific device
    callFunction(sFunction, sDeviceName, dsDpes, bForHistory);
	} else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateDPEs: Function %s does not exists", sFunction);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 105, sErrMsg);
    throw( error );	
  }

}

/**
  Returns the main color associated with the device state. This color is based on 
  the dpes listed in the vacuumExtension_StateDPEs for the dedicated device type

  @param string sDeviceName, input,  device dp name
  @param dyn_string dsDpes, input,  the list of dpe names used to determine the color
  @param dyn_anytype daValues, input,  the list of dpe values used to determine the color
  @param string &sColor, output, main color
  @param string &sState, main state string
  
  @return The main color corresponding to the state of the object
*/

string vacuumExtension_StateMainColorAndString(string sDeviceName, dyn_string dsDpes, dyn_anytype daValues, string &sColor, string &sState){
  // Get the device name from the 
  string sDeviceType = dpTypeName(sDeviceName), sFunction = "";
  if(sDeviceType != ""){
    sFunction = sDeviceType +"_ColorAndString";
  } else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateMainColorAndString: Device with name %s does not exists", sDeviceName);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 104, sErrMsg);
    throw( error );	
  }
  
  // Exec device type specific function if available
  if (isFunctionDefined(sFunction) && sFunction != "") {
    callFunction(sFunction, sDeviceName, dsDpes, daValues, sColor, sState, true);
 	} else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateMainColorAndString: Function %s does not exists", sFunction);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 105, sErrMsg);
    throw( error );	
  }
  
  return sColor;
}

/**
  Returns:
        main color associated with the device state. This color is based on device input dpes and the global bool realtime alert from unicos manager
        fill/empty behavior of the bar

  @param string sDeviceName, input,  device dp name
  @param string &sColor, output, main color
  @param bool &bFill, output, fill(true) or empty(false) bar behaviour

  @return The main color corresponding to the state of the object
*/

string vacuumExtension_StateMainColorFillAndString(string sDeviceName, dyn_string dsDpes, dyn_anytype daValues, string &sColor, bool &bFill, string &sState) {
  // Get the device name from the
  string sDeviceType = dpTypeName(sDeviceName), sFunction = "";
  if(sDeviceType != ""){
    sFunction = sDeviceType +"_ColorFillAndString";
  } else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateMainColorFillAndString: Device with name %s does not exists", sDeviceName);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 106, sErrMsg);
    throw(error);
  }

  // Exec device type specific function if available
  if (isFunctionDefined(sFunction) && sFunction != "") {
    // Execute function for the specific device
    callFunction(sFunction, sDeviceName, dsDpes, daValues, sColor, bFill, sState);
  } else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateMainColorFillAndString: Function %s does not exists", sFunction);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 107, sErrMsg);
    throw(error);
  }

  return sColor;
}


/**
  Returns the error message associated with the device state. This string is based on 
  the dpes listed in the vacuumExtension_StateDPEs for the dedicated device type. dpe This list should
  therefore contain the error code status if neccesary. Unsupported devices will 
  return an empty string

  @param dyn_string dsDpes The list of dpe names used to determine the color
  @param dyn_anytype daValues The list of dpe values used to determine the color
  
  @return The main state string corresponding to the state of the object
*/  

string vacuumExtension_StateErrorMessage(string sDeviceName, dyn_string dsDpes, dyn_anytype daValues){
  // Get the device name from the 
  string sDeviceType = dpTypeName(sDeviceName), sFunction = "", sMessage;
  if(sDeviceType != ""){
    sFunction = sDeviceType +"_StateErrorMessage";
  } else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateErrorMessage: Device with name %s does not exists", sDeviceName);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 104, sErrMsg);
    throw( error );	
  }
  
  // Exec device type specific function if available
  if (isFunctionDefined(sFunction) && sFunction != "") {
    // Execute callback function for the specific device
    sMessage = callFunction(sFunction, dsDpes, daValues);
 	} else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateErrorMessage: Function %s does not exists", sFunction);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 105, sErrMsg);
    throw( error );	
  }
  return sMessage;
    
}


/**
  Returns the warning message associated with the device state. This string is based on 
  the dpes listed in the vacuumExtension_StateDPEs for the dedicated device type. This dpe list should
  therefore contain the warning code status if neccesary. Unsupported devices will 
  return an empty string

  @param dyn_string dsDpes The list of dpe names used to determine the color
  @param dyn_anytype daValues The list of dpe values used to determine the color
  
  @return The main state string corresponding to the state of the object
*/

string vacuumExtension_StateWarningMessage(string sDeviceName, dyn_string dsDpes, dyn_anytype daValues){
  // Get the device name from the 
  string sDeviceType = dpTypeName(sDeviceName), sFunction = "", sMessage;
  if(sDeviceType != ""){
    sFunction = sDeviceType +"_StateWarningMessage";
  } else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateWarningMessage: Device with name %s does not exists", sDeviceName);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 104, sErrMsg);
    throw( error );	
  }
  
  // Exec device type specific function if available
  if (isFunctionDefined(sFunction) && sFunction != "") {
    // Execute callback function for the specific device
    sMessage = callFunction(sFunction, dsDpes, daValues);
 	} else {
    string sErrMsg;
    sprintf(sErrMsg, "vacuumExtension_StateWarningMessage: Function %s does not exists", sFunction);
    errClass error = makeError("unVacuum", PRIO_SEVERE, ERR_IMPL, 105, sErrMsg);
    throw( error );	
  }
  return sMessage;
}

/**
  Extracts the datapoint type name from the DPE list
  
  @argument dyn_string dsDpes The list of DPEs
  @return The datapoint type name
*/
string vacuumExtension_ExtractDeviceType(dyn_string dsDpes){
  if(dynlen(dsDpes) >= 3){
    return dpTypeName(dsDpes[3]);    
  }
}


/**
  Lists the dpes required for building the state of the device
  
  @param string sDeviceName The name of the device which is in UNICOS the same as the dpname
  @param dyn_string dsDpes The list of datapoints elements required for building the state
  @param bool bForHistory The DPEs will be used for state history, i.e. only DPEs with _archive
                            config are of interest
  
*/
CPC_ProcessControlObject_StateDPEs(string sDeviceName, dyn_string &dsDpes, bool bForHistory = true) {
    dynAppend(dsDpes, sDeviceName + ".eventMask");                                  // 3
    
    if(bForHistory) {
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg01");                    // 4
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg01:_online.._invalid");  // 5
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg02");                    // 6
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg02:_online.._invalid");  // 7
    }
    else {
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg01");                      // 4
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg01:_online.._invalid");    // 5
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg02");                      // 6
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg02:_online.._invalid");    // 7
    }
    
    dynAppend(dsDpes, sDeviceName + ".ProcessInput.OpMoSt");                             // 8
    dynAppend(dsDpes, sDeviceName + ".ProcessInput.OpMoSt:_online.._invalid");           // 9
    dynAppend(dsDpes, sDeviceName + ".ProcessInput.AuOpMo:_online.._invalid");           // 10
    dynAppend(dsDpes, sDeviceName + ".disabled");                                     // 11
    
    if (g_bBoolAlert) {
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.StartISt:_alert_hdl.._act_state");  // 12
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.TStopISt:_alert_hdl.._act_state");  // 13
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.FuStopISt:_alert_hdl.._act_state"); // 14
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.AlSt:_alert_hdl.._act_state");      // 15
    }
}

/**
  Returns the main color associated with the device state. This color is based on 
  the dpes listed in the CPC_ProcessControlObject_StateDPEs for the dedicated device type. 
  In case of the state history the EvStsReg## is used because this is archived, while for determining 
  the current state the StsReg## is used.

  @param dyn_string dsDpes The list of dpe names used to determine the color
  @param dyn_anytype daValues The list of dpe values used to determine the color
  
  Datapoint elements and related values
  1  int	    frontend.alarm
  2  bool 		frontend.enabled
  3  bool     device.eventMask
  4  bit32		device.ProcessInput.(Ev)StsReg01
  5  bool 		device.ProcessInput.(Ev)StsReg01:_online.._invalid
  6  bit32		device.ProcessInput.(Ev)StsReg02
  7  bool		  device.ProcessInput.(Ev)StsReg02:_online.._invalid
  8  int		  device.ProcessInput.OpMoSt
  9  int		  device.ProcessInput.OpMoSt:_online.._invalid 
  10 int      device.ProcessInput.AuOpMo:_online.._invalid
  11 bool		  device.disabled
  12 int		  device.ProcessInput.StartISt:_alert_hdl.._act_state
  13 int		  device.ProcessInput.TStopISt:_alert_hdl.._act_state
  14 int		  device.ProcessInput.FuStopISt:_alert_hdl.._act_state
  15 int		  device.ProcessInput.AlSt:_alert_hdl.._act_state	
  
  @param string sColor, output, the main color string of device
  @param string sString, output, the main state string of device
  @param bool bUseOffColor, input, Off color is return in &sColor when device is Off, false in case of WidgetColor
  @param bool bBoolAlert, input, realtime alert according to UNICOS alert manager

  @param bool bIsRestartNeeded, input, there is a restart needed according to UNICOS device manager
  
  @see For the stsReg bit meaning check the specifications file: https://svn.cern.ch/reps/en-ice-svn/trunk/applications/vacuum/UNICOS_VAC_DEVICES/Design/Object_ManReg_StsReg_Mapping.xls
  @see Other usufull documentation: https://edms.cern.ch/ui/file/1228441/1.6.0/UCPC6_RP1.6.0_Manual.pdf  
  
  @return '0' when function proceed successfully, '-1' when function proceed with error
  
*/

int CPC_ProcessControlObject_ColorAndString(string sDeviceName, dyn_string dsDpes, dyn_anytype daValues, string &sColor,
                               string &sState, bool bUseOffColor, bool bBoolAlert = false,
                               bool bIsRestartNeeded = false){
  // INIT VARIABLES
  int iSystemIntegrityAlarmValue = daValues[1];
  bool bSystemIntegrityAlarmEnabled = daValues[2];
  bit32 bit32StsReg01 = daValues[4];
  bit32 bit32StsReg02 = daValues[6];
  unsigned iOpMoSt = daValues[8];  
  bool bDisabled = daValues[11];
    
  int iStartIStState, iTStopIStState, iFuStopIStState, iAlSt;
  if (bBoolAlert) {
    iStartIStState = daValues[12];
    iTStopIStState = daValues[13];
    iFuStopIStState = daValues[14];
    iAlSt = daValues[15];
  }
  
  // Determine valid state
  bool bInvalid = daValues[5] || daValues[7] || daValues[9] || daValues[10];
    
  // --- Get main state string
  //--------------------------
  if (sDeviceName!="") {
     CPC_ProcessControlObject_getOptionModesLabel(sDeviceName, iOpMoSt, sState);
  }
  sState = (State == "") ? "Unknown" : sState;

  // --- Get main color string
    
  // Determine alarm color booleans
  bool bAlarm = getBit(bit32StsReg01, CPC_StsReg01_TSTOPIST) || getBit(bit32StsReg02, CPC_StsReg02_FUSTOPIST) || getBit(bit32StsReg01, CPC_StsReg01_ALST);
  bool bAlarmUnack= getBit(bit32StsReg01, CPC_StsReg01_ALUNACK) == 1 &&
             (iTStopIStState == 1 || 
              iTStopIStState == 3 || 
              iFuStopIStState == 1 || 
              iFuStopIStState == 3);
  
  // Warning boolean if Opened with Error(ObjectSt == 0) or IO Error 
  bool bWarning = getBit(bit32StsReg01, CPC_StsReg01_IOERRORW) == 1;

  // Mode boolean
  bool bForcedMode = getBit(bit32StsReg01, CPC_StsReg01_FoMoSt) == 1;  
  // Valve is open in Opened with Error (ObjectSt == 0) and Open without error(ObjectSt == 1)
  bool bOn = getBit(bit32StsReg01, CPC_StsReg01_ONST) == 1; 
  // Show half animation when the PCO is in undefined state
  bool bHalf = getBit(bit32StsReg01, CPC_StsReg01_ONST) == getBit(bit32StsReg01, CPC_StsReg01_OFFST); 
  
  // Create boolean array for corresponding colors
  dyn_bool dbBodyStateBool = makeDynBool(
      bInvalid,              // Invalid color
      bAlarmUnack,           // Alarm Unacknowledged color
      bAlarm,                // Alarm active color
      bDisabled,             // Device disabled color
      bWarning,              // Warning color
      bForcedMode,           // Forced mode color
      bUseOffColor && bHalf, // Undefined color
      !bOn && bUseOffColor,  // Off color (optional color)
      false,                 // Under Range color
      bOn                    // OK color
  );
    
  // Create boolen array for corresponding state by using modified version of boolean array   
  unVacObject_WidgetBoolMapping(dbBodyStateBool, CPC_BODY_COLORS, sColor);
  
  return 0;
}



/**
  Lists the dpes required for building the state of the device
  
  @param string sDeviceName The name of the device which is in UNICOS the same as the dpname
  @param dyn_string dsDpes The list of datapoints elements required for building the state
  @param bool bForHistory The DPEs will be used for state history, i.e. only DPEs with _archive
                            config are of interest
  
*/

CPC_OnOff_StateDPEs(string sDeviceName, dyn_string &dsDpes, bool bForHistory = true) {
    dynAppend(dsDpes, sDeviceName + ".eventMask");                                  // 3
    
    if(bForHistory) {
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg01");                    // 4
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg01:_online.._invalid");  // 5
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg02");                    // 6
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.evStsReg02:_online.._invalid");  // 7
    }
    else {
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg01");                      // 4
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg01:_online.._invalid");    // 5
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg02");                      // 6
      dynAppend(dsDpes, sDeviceName + ".ProcessInput.StsReg02:_online.._invalid");    // 7
    }
    
    if (g_bBoolAlert) {
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.StartISt:_alert_hdl.._act_state");  // 12
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.TStopISt:_alert_hdl.._act_state");  // 13
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.FuStopISt:_alert_hdl.._act_state"); // 14
        dynAppend(dsDpes, sDeviceName + ".ProcessInput.AlSt:_alert_hdl.._act_state");      // 15
    }
}

/**
  Returns the main color associated with the device state. This color is based on 
  the dpes listed in the CPC_OnOff_StateDPEs for the dedicated device type. 
  In case of the state history the EvStsReg## is used because this is archived, while for determining 
  the current state the StsReg## is used.

  @param dyn_string dsDpes The list of dpe names used to determine the color
  @param dyn_anytype daValues The list of dpe values used to determine the color
  
  Datapoint elements and related values
  1  int	    frontend.alarm
  2  bool 		frontend.enabled
  3  bool     device.eventMask
  4  bit32		device.ProcessInput.(Ev)StsReg01
  5  bool 		device.ProcessInput.(Ev)StsReg01:_online.._invalid
  6  bit32		device.ProcessInput.(Ev)StsReg02
  7  bool		  device.ProcessInput.(Ev)StsReg02:_online.._invalid
  8  int		  device.ProcessInput.StartISt:_alert_hdl.._act_state
  9  int		  device.ProcessInput.TStopISt:_alert_hdl.._act_state
  10 int		  device.ProcessInput.FuStopISt:_alert_hdl.._act_state
  11 int		  device.ProcessInput.AlSt:_alert_hdl.._act_state	
  
  @param string sColor, output, the main color string of device
  @param string sString, output, the main state string of device
  @param bool bUseOffColor, input, Off color is return in &sColor when device is Off, false in case of WidgetColor
  @param bool bBoolAlert, input, realtime alert according to UNICOS alert manager

  @param bool bIsRestartNeeded, input, there is a restart needed according to UNICOS device manager
  
  @see For the stsReg bit meaning check the specifications file: https://svn.cern.ch/reps/en-ice-svn/trunk/applications/vacuum/UNICOS_VAC_DEVICES/Design/Object_ManReg_StsReg_Mapping.xls
  @see Other usufull documentation: https://edms.cern.ch/ui/file/1228441/1.6.0/UCPC6_RP1.6.0_Manual.pdf  
  
  @return '0' when function proceed successfully, '-1' when function proceed with error
  
*/

int CPC_OnOff_ColorAndString(string sDeviceName, dyn_string dsDpes, dyn_anytype daValues, string &sColor,
                               string &sState, bool bUseOffColor, bool bBoolAlert = false,
                               bool bIsRestartNeeded = false){
  // INIT VARIABLES
  int iSystemIntegrityAlarmValue = daValues[1];
  bool bSystemIntegrityAlarmEnabled = daValues[2];
  bit32 bit32StsReg01 = daValues[4];
  bit32 bit32StsReg02 = daValues[6];
    
  int iStartIStState, iTStopIStState, iFuStopIStState, iAlSt;
  if (bBoolAlert) {
    iStartIStState = daValues[8];
    iTStopIStState = daValues[9];
    iFuStopIStState = daValues[10];
    iAlSt = daValues[11];
  }
  
  // Determine valid state
  bool bInvalid = daValues[5] || daValues[7];

  // --- Get main color string
    
  // Determine alarm color booleans
  bool bAlarm = getBit(bit32StsReg01, CPC_StsReg01_TSTOPIST) || getBit(bit32StsReg02, CPC_StsReg02_FUSTOPIST) || getBit(bit32StsReg01, CPC_StsReg01_ALST);
  bool bAlarmUnack= getBit(bit32StsReg01, CPC_StsReg01_ALUNACK) == 1 &&
             (iTStopIStState == 1 || 
              iTStopIStState == 3 || 
              iFuStopIStState == 1 || 
              iFuStopIStState == 3);
  
  // Warning boolean if Opened with Error(ObjectSt == 0) or IO Error 
  bool bWarning = getBit(bit32StsReg01, CPC_StsReg01_IOERRORW) == 1;

  // Mode boolean
  bool bForcedMode = getBit(bit32StsReg01, CPC_StsReg01_FoMoSt) == 1;  
  // Valve is open in Opened with Error (ObjectSt == 0) and Open without error(ObjectSt == 1)
  bool bOn = getBit(bit32StsReg01, CPC_StsReg01_ONST) == 1; 
  // Show half animation when the PCO is in undefined state
  bool bHalf = getBit(bit32StsReg01, CPC_StsReg01_ONST) == getBit(bit32StsReg01, CPC_StsReg01_OFFST); 
  
  // Create boolean array for corresponding colors
  dyn_bool dbBodyStateBool = makeDynBool(
      bInvalid,              // Invalid color
      bAlarmUnack,           // Alarm Unacknowledged color
      bAlarm,                // Alarm active color
      false,             // Device disabled color
      bWarning,              // Warning color
      bForcedMode,           // Forced mode color
      bUseOffColor && bHalf, // Undefined color
      !bOn && bUseOffColor,  // Off color (optional color)
      false,                 // Under Range color
      bOn                    // OK color
  );
    
  // Create boolen array for corresponding state by using modified version of boolean array   
  unVacObject_WidgetBoolMapping(dbBodyStateBool, CPC_BODY_COLORS, sColor);
  
  // --- Get main state string
  if(bHalf){
    sState = "Undefined";
  } else if(bOn){
    sState = "On/Open";
  } else {
    sState = "Off/Closed";
  }  
  
  return 0;
}

